/*
 * QuadTree.cpp
 *
 *  Created on: Feb 5, 2009
 *      Author: sri
 */
#include <iostream>
#include <limits>
using namespace std;
#include "QuadTree.h"
#include <cmath>


QuadTree::QuadTree() {
	x1=x2=y1=y2=0;
	currentDepth=0;
	parent=NULL;
	leftWallOwner=this;
	rightWallOwner=this;
	topWallOwner=this;
	bottomWallOwner=this;
	quadrant=0;

	mass=0;
	cm_x=cm_y=0;
}

QuadTree* QuadTree::getParent(){
	return parent;
}

QuadTree* QuadTree::getChild(int index){
	try{
		return children.at(index);
	}catch(std::out_of_range outOfRange){
		return NULL;
	}
}

bool QuadTree::addChild(QuadTree* child){
	try{
		children.push_back(child);
		child->parent=this;
		child->currentDepth=currentDepth+1;
		return true;
	}catch (std::bad_alloc badAlloc){
		return false;
	}
}

void QuadTree::deleteSelf(bool recursive){
	if (recursive){
		for (unsigned int i=0;i<children.size();i++){
			if (children[i])
				children[i]->deleteSelf(true);
			delete children[i];
		}
	}
	data.clearList();
	children.clear();
}

void QuadTree::setQuadrant(unsigned char quadrant){
//Quadrant 1 is the first quadrant: top right corner
//Note that (0,0) is on top left corner, with X increasing
//towards the right side, and Y increasing downwards.
	if (quadrant==1 || quadrant==2){
		y1=parent->y1;
		y2=(y1+parent->y2)/2;
		topWallOwner=parent->topWallOwner;
		bottomWallOwner=this;
	}
	if (quadrant==3 || quadrant==4){
		y2=parent->y2;
		y1=(y2+parent->y1)/2;
		topWallOwner=this;
		bottomWallOwner=parent->bottomWallOwner;
	}
	if (quadrant==1 || quadrant==4){
		x2=parent->x2;
		x1=(x2+parent->x1)/2;
		leftWallOwner=this;
		rightWallOwner=parent->rightWallOwner;
	}
	if (quadrant==2 || quadrant==3){
		x1=parent->x1;
		x2=(x1+parent->x2)/2;
		leftWallOwner=parent->leftWallOwner;
		rightWallOwner=this;
	}
	this->quadrant=quadrant;
}

void QuadTree::setBounds(float x1, float x2, float y1, float y2){
	this->x1=x1;
	this->x2=x2;
	this->y1=y1;
	this->y2=y2;
	propogateBounds();
}

void QuadTree::propogateBounds(){
	if (!isLeafNode()){
		children[0]->setQuadrant(1);
		children[1]->setQuadrant(2);
		children[2]->setQuadrant(3);
		children[3]->setQuadrant(4);
	}
}

bool QuadTree::boundsTest(const Point& p) const{
	if (p.x>x1 && p.x<=x2 && p.y>y1 && p.y<=y2)
		return true;
	else
		return false;
}

QuadTree* QuadTree::traverseTree(const Point& p,unsigned int levels){
	QuadTree* tmp;
	if (boundsTest(p)==true){
		if (currentDepth<levels && !isLeafNode()){//not leaf node
			for (int i=0;i<4;i++){
				if ((tmp=children[i]->traverseTree(p,levels))){
					return tmp;
				}
			}
		}else{//leaf node
			return this;
		}
	}
	return NULL;
}

void QuadTree::splitNode(){
	if(!isLeafNode())
		return;//Cannot split. already has children

	for (int i=0;i<4;i++){
		addChild(new QuadTree);
	}
	propogateBounds();//initialize the boundaries of the children
	propogateData();//move the data from this node to children

	for (int i=0;i<4;i++){
		if (children[i]->getNumPoints()>maxPoints && children[i]->currentDepth<maxDepth){
			children[i]->splitNode();
		}
	}
}

void QuadTree::propogateData(){
	list<Point>::iterator iterate=data.getBegin();
	for (;iterate!=data.getEnd();iterate++){
		for (int j=0;j<4;j++){
			if (children[j]->boundsTest(*iterate)){
				children[j]->data.addPoint(*iterate);
				break;
			}
		}
	}
	data.clearList();
}

void QuadTree::recombine(){
	if (isLeafNode())
		return;//cannot recombine. no children
	for (int i=0;i<4;i++){
		children[i]->recombine();//FIXME: Can make this more efficient
		data.points.insert(data.points.end(),children[i]->data.points.begin(),children[i]->data.points.end());
		children[i]->deleteSelf();
	}
}

QuadTree* QuadTree::addPoint(const Point& p){
	if (isLeafNode()){
		data.addPoint(p);
		if (data.getNumPoints()>maxPoints && currentDepth<maxDepth){
			splitNode();
			return traverseTree(p);
		}else{
			return this;
		}
	}else{
		QuadTree* target=traverseTree(p);
		if (target)
			target->addPoint(p);
		return target;
	}
}

unsigned int QuadTree::getNumPoints(){
	return data.getNumPoints();
}

PointList* QuadTree::getPointList(){
	return &data;
}

bool QuadTree::isLeafNode(){
	return (children.size()==0);
}

void QuadTree::collectChildren(vector<QuadTree*>& storage, unsigned char quadrants, unsigned int minDepth, unsigned int maxDepth){
	if (isLeafNode()){
		if (currentDepth>=minDepth) {
			storage.push_back(this);
			//cout<<"Depth: "<<currentDepth <<", Min: "<<minDepth<<", Max: " << maxDepth<<endl;
		}
	}else if (currentDepth<maxDepth){
		if (quadrants & Q_ONE) children[0]->collectChildren(storage,quadrants,minDepth,maxDepth);
		if (quadrants & Q_TWO) children[1]->collectChildren(storage,quadrants,minDepth,maxDepth);
		if (quadrants & Q_THREE) children[2]->collectChildren(storage,quadrants,minDepth,maxDepth);
		if (quadrants & Q_FOUR) children[3]->collectChildren(storage,quadrants,minDepth,maxDepth);
	}
}

vector<QuadTree*> QuadTree::getLeftNeighbours(unsigned int far_depth){
	vector<QuadTree*> output;
	if (leftWallOwner->parent==NULL){
		return output;//the root node is the left wall owner. no neighbours
	}else{
		Point p((3*x1-x2)/2,(y1+y2)/2);
		QuadTree* neighbour=leftWallOwner->parent->traverseTree(p,currentDepth);
		//neighbour->collectRightMostChildren(output);
		if(far_depth==0){
			neighbour->collectChildren(output,Q_ONE | Q_FOUR);
		}else{
			neighbour->collectChildren(output,Q_ONE | Q_TWO | Q_THREE | Q_FOUR, far_depth, far_depth);
		}
		//output.push_back(neighbour);
	}
	return output;
}

vector<QuadTree*> QuadTree::getRightNeighbours(unsigned int far_depth){
	vector<QuadTree*> output;
	if (rightWallOwner->parent==NULL){
		return output;//the root node is the right wall owner. no neighbours
	}else{
		Point p((3*x2-x1)/2,(y1+y2)/2);
		QuadTree* neighbour=rightWallOwner->parent->traverseTree(p,currentDepth);
		//neighbour->collectLeftMostChildren(output);
		if(far_depth==0){
			neighbour->collectChildren(output,Q_TWO | Q_THREE);
		}else{
			neighbour->collectChildren(output,Q_ONE | Q_TWO | Q_THREE | Q_FOUR, far_depth, far_depth);
		}
		//output.push_back(neighbour);
	}
	return output;
}

vector<QuadTree*> QuadTree::getTopNeighbours(unsigned int far_depth){
	vector<QuadTree*> output;
	if (topWallOwner->parent==NULL){
		return output;//the root node is the left wall owner. no neighbours
	}else{
		Point p((x1+x2)/2,(3*y1-y2)/2);
		QuadTree* neighbour=topWallOwner->parent->traverseTree(p,currentDepth);
		//neighbour->collectBottomMostChildren(output);
		if(far_depth==0){
			neighbour->collectChildren(output,Q_THREE | Q_FOUR);
		}else{
			neighbour->collectChildren(output,Q_ONE | Q_TWO | Q_THREE | Q_FOUR, far_depth, far_depth);
		}
		//output.push_back(neighbour);
	}
	return output;
}

vector<QuadTree*> QuadTree::getBottomNeighbours(unsigned int far_depth){
	vector<QuadTree*> output;
	if (bottomWallOwner->parent==NULL){
		return output;//the root node is the left wall owner. no neighbours
	}else{
		Point p((x1+x2)/2,(3*y2-y1)/2);
		QuadTree* neighbour=bottomWallOwner->parent->traverseTree(p,currentDepth);
		//neighbour->collectTopMostChildren(output);
		if(far_depth==0){
			neighbour->collectChildren(output,Q_ONE | Q_TWO);
		}else{
			neighbour->collectChildren(output,Q_ONE | Q_TWO | Q_THREE | Q_FOUR, far_depth, far_depth);
		}
		//output.push_back(neighbour);
	}
	return output;
}

vector<QuadTree*> QuadTree::getNE_Neighbour(unsigned int far_depth){
	vector<QuadTree*> output;
	QuadTree* wallOwner;

	wallOwner=(topWallOwner->currentDepth<rightWallOwner->currentDepth)?topWallOwner:rightWallOwner;

	if (wallOwner->parent==NULL){
		return output;
	}else{
		Point p((3*x2-x1)/2,(3*y1-y2)/2);
		QuadTree* neighbour=wallOwner->parent->traverseTree(p,currentDepth);
		//neighbour->collectTopMostChildren(output);
		if(far_depth==0){
			if (neighbour) neighbour->collectChildren(output,Q_THREE);
		}else{
			if (neighbour) neighbour->collectChildren(output,Q_ONE | Q_TWO | Q_THREE | Q_FOUR,far_depth,far_depth);
		}
		//output.push_back(neighbour);
	}

	return output;
}

vector<QuadTree*> QuadTree::getNW_Neighbour(unsigned int far_depth){
	vector<QuadTree*> output;
	QuadTree* wallOwner;

	//choose topmost owner
	wallOwner=(topWallOwner->currentDepth<leftWallOwner->currentDepth)?topWallOwner:leftWallOwner;
	if (wallOwner->parent==NULL){
		return output;
	}else{
		Point p((3*x1-x2)/2,(3*y1-y2)/2);
		QuadTree* neighbour=wallOwner->parent->traverseTree(p,currentDepth);
		//neighbour->collectTopMostChildren(output);

		if(far_depth==0){
			if (neighbour) neighbour->collectChildren(output,Q_FOUR);
		}else{
			if (neighbour) neighbour->collectChildren(output,Q_ONE | Q_TWO | Q_THREE | Q_FOUR,far_depth,far_depth);
		}
		//output.push_back(neighbour);
	}
	return output;
}

vector<QuadTree*> QuadTree::getSE_Neighbour(unsigned int far_depth){
	vector<QuadTree*> output;
	QuadTree* wallOwner;

	wallOwner=(bottomWallOwner->currentDepth<rightWallOwner->currentDepth)?bottomWallOwner:rightWallOwner;

	if (wallOwner->parent==NULL){
		return output;
	}else{
		Point p((3*x2-x1)/2,(3*y2-y1)/2);
		QuadTree* neighbour=wallOwner->parent->traverseTree(p,currentDepth);
		//neighbour->collectTopMostChildren(output);
		if(far_depth==0){
			if (neighbour) neighbour->collectChildren(output,Q_TWO);
		}else{
			if (neighbour) neighbour->collectChildren(output,Q_ONE | Q_TWO | Q_THREE | Q_FOUR, far_depth,far_depth);
		}
		//output.push_back(neighbour);
	}
	return output;
}

vector<QuadTree*> QuadTree::getSW_Neighbour(unsigned int far_depth){
	vector<QuadTree*> output;
	QuadTree* wallOwner;

	wallOwner=(bottomWallOwner->currentDepth<leftWallOwner->currentDepth)?bottomWallOwner:leftWallOwner;

	if (wallOwner->parent==NULL){
		return output;
	}else{
		Point p((3*x1-x2)/2,(3*y2-y1)/2);
		QuadTree* neighbour=wallOwner->parent->traverseTree(p,currentDepth);
		//neighbour->collectTopMostChildren(output);
		if(far_depth==0){
			if (neighbour) neighbour->collectChildren(output,Q_ONE);
		}else{
			//cout<<"SouthWest collection: " << far_depth<<endl;
			if (neighbour) neighbour->collectChildren(output,Q_ONE | Q_TWO | Q_THREE | Q_FOUR, far_depth,far_depth);
		}

		//output.push_back(neighbour);
	}
	return output;
}

vector<QuadTree*> QuadTree::getNearNeighbours(){
	vector<QuadTree*> output;
	vector<QuadTree*> temp;

	temp=getLeftNeighbours();
	output.insert(output.end(),temp.begin(),temp.end());
	temp.clear();

	temp=getRightNeighbours();
	output.insert(output.end(),temp.begin(),temp.end());
	temp.clear();

	temp=getTopNeighbours();
	output.insert(output.end(),temp.begin(),temp.end());
	temp.clear();

	temp=getBottomNeighbours();
	output.insert(output.end(),temp.begin(),temp.end());
	temp.clear();

	temp=getNE_Neighbour();
	if (temp.size()>0) output.push_back(temp[0]);

	temp=getNW_Neighbour();
	if (temp.size()>0) output.push_back(temp[0]);

	temp=getSE_Neighbour();
	if (temp.size()>0) output.push_back(temp[0]);

	temp=getSW_Neighbour();
	if (temp.size()>0) output.push_back(temp[0]);

	return output;
}
vector<QuadTree*> QuadTree::getFarNeighbours(){
	vector<QuadTree*> output;
	vector<QuadTree*> temp;

	if (parent==NULL){//root node has no parent
		return output;
	}

	temp=parent->getLeftNeighbours(currentDepth);
	output.insert(output.end(),temp.begin(),temp.end());
	temp.clear();

	temp=parent->getRightNeighbours(currentDepth);
	output.insert(output.end(),temp.begin(),temp.end());
	temp.clear();

	temp=parent->getTopNeighbours(currentDepth);
	output.insert(output.end(),temp.begin(),temp.end());
	temp.clear();

	temp=parent->getBottomNeighbours(currentDepth);
	output.insert(output.end(),temp.begin(),temp.end());
	temp.clear();

	temp=parent->getNE_Neighbour(currentDepth);
	output.insert(output.end(),temp.begin(),temp.end());
	temp.clear();

	temp=parent->getNW_Neighbour(currentDepth);
	output.insert(output.end(),temp.begin(),temp.end());
	temp.clear();

	temp=parent->getSE_Neighbour(currentDepth);
	output.insert(output.end(),temp.begin(),temp.end());
	temp.clear();

	temp=parent->getSW_Neighbour(currentDepth);
	output.insert(output.end(),temp.begin(),temp.end());
	temp.clear();

	return output;
}

QuadTree * QuadTree::rayTrace(const VectorLine& v, float threshold, float* distance, float corner){
	float rayHit=hitTest(v, threshold);
	bool grazing=false;

	if (rayHit<0)
		return NULL;
	else{

		//check for grazing rays (they hit the node at outer corners)
		//if (!(v.dx==0 || v.dy==0)){//dont do this for horizontal/vertical lines.(A hack)
			float incomingX=v.evaluateX(rayHit-VectorLine::delta);
			float incomingY=v.evaluateY(rayHit-VectorLine::delta);
			float my_corner=corner*(x2-x1);//scale the corner threshold as per size of this node.
			if ((incomingX<x1+my_corner || incomingX>x2-my_corner) && (incomingY<y1+my_corner || incomingY>y2-my_corner)){
//				cout<<v.perpendicularDistance(getCenterX(),getCenterY())<<","<<(x2-x1)/2<<endl;
				if (v.perpendicularDistance(getCenterX(),getCenterY()) > (x2-x1)/2){//not intersect the inscribed circle
					grazing=true;
				}
			}
		//}

		if (grazing || getNumPoints()==0 || boundsTest(Point(v.px,v.py))){
			//if ray origin is inside this box or if box is empty, get the next box along the ray's path
			Point extendRay=Point(v.evaluateX(rayHit+VectorLine::delta),v.evaluateY(rayHit+VectorLine::delta));
			vector<QuadTree*> nearCells=getNearNeighbours();
			for (unsigned int i=0;i<nearCells.size();i++){//FIXME: this test need not be done for neighbours on all 4 sides
				if (nearCells[i]->boundsTest(extendRay)){
					return nearCells[i]->rayTrace(v,rayHit+VectorLine::delta, distance, corner);
				}
			}
			return NULL;//none of the neighbours intersect the ray
		}else{
			//otherwise, this is the node we are looking for
			if (distance!=NULL) *distance=threshold;
			return this;
		}
	}
}

vector<QuadTree*> QuadTree::drawLine(const VectorLine& v, float threshold, float corner){
	vector<QuadTree*> retval;
	VectorLine ray(v);
	float distance;
	QuadTree* currentLeaf=this;
	while (currentLeaf){
		retval.push_back(currentLeaf);//add this leaf to return list

		currentLeaf=currentLeaf->rayTrace(ray,0,&distance,corner);//get next leaf
		distance+=VectorLine::delta;
		ray.px=ray.evaluateX(distance);//move the ray to next leaf
		ray.py=ray.evaluateY(distance);

		//cout<<currentLeaf<<" "<<distance<<endl;
	}
	return retval;
}

float QuadTree::hitTest(const VectorLine& v, float threshold){
	static float t1,t2,t3,t4;
	static float xHit1,xHit2,yHit1,yHit2;
	static float smallT;
	static float maxFloat=numeric_limits<float>::max();

	t1=v.hIntercept(y1);
	t2=v.hIntercept(y2);
	t3=v.vIntercept(x1);
	t4=v.vIntercept(x2);

	//ignore values below a threshold.
	//The threshold is the min distance from the ray origin that the intercept must occur
	xHit1 = (t1<threshold)?maxFloat:v.evaluateX(t1);
	xHit2 = (t2<threshold)?maxFloat:v.evaluateX(t2);
	yHit1 = (t3<threshold)?maxFloat:v.evaluateY(t3);
	yHit2 = (t4<threshold)?maxFloat:v.evaluateY(t4);

	//ignore values outside the box
	if (xHit1<=x1 || xHit1>x2){
		t1=maxFloat;
	}
	if (xHit2<=x1 || xHit2>x2){
		t2=maxFloat;
	}
	if (yHit1<=y1 || yHit1>y2){
		t3=maxFloat;
	}
	if (yHit2<=y1 || yHit2>y2){
		t4=maxFloat;
	}

	//get smallest +ve t value
	smallT=min(min(t1,t2),min(t3,t4));
	if (smallT==maxFloat){
		return -1;//No intersections
	}
	return smallT;
}

float QuadTree::getX1(){
	return x1;
}
float QuadTree::getX2(){
	return x2;
}
float QuadTree::getY1(){
	return y1;
}
float QuadTree::getY2(){
	return y2;
}

float QuadTree::getCenterX(){
	return (x1+x2)/2;
}
float QuadTree::getCenterY(){
	return (y1+y2)/2;
}

void QuadTree::setMaxDepth(unsigned int d){
	maxDepth=d;
}

void QuadTree::setMaxPointsPerLeaf(unsigned int p){
	maxPoints=p;
}

unsigned int QuadTree::getMaxDepth(){
	return maxDepth;
}
unsigned int QuadTree::getMaxPointsPerLeaf(){
	return maxPoints;
}

void QuadTree::collectPoints(PointList& p){
	if (isLeafNode()){
		p.append(data);
	}else{
		for (int i=0;i<4;i++){
			children[i]->collectPoints(p);
		}
	}
}

///////////////////////////////////////////////////////////
void QuadTree::calculateCM(){
	static list<Point>::iterator iterate;

	cm_x=cm_y=0.0f;
	if (isLeafNode()){
		mass=data.getNumPoints();
		iterate=data.getBegin();
		for(;iterate!=data.getEnd();iterate++){
			cm_x+=iterate->x;
			cm_y+=iterate->y;
		}
		cm_x/=(float)mass;
		cm_y/=(float)mass;
	}else{
		mass=0;
		for (unsigned int i=0;i<4;i++){
			children[i]->calculateCM();
			mass+=children[i]->mass;
			cm_x+=children[i]->mass*children[i]->cm_x;
			cm_y+=children[i]->mass*children[i]->cm_y;
		}
		cm_x/=(float)mass;
		cm_y/=(float)mass;
	}
}
void QuadTree::gatherForces(Point& p){
	static float distance,distance2;
	static list<Point>::iterator iterate;

	distance2=(cm_x-p.x)*(cm_x-p.x) + (cm_y-p.y)*(cm_y-p.y);
	distance=sqrt(distance2);

	if (distance>SIMULATION_ACCURACY && distance>0.001){
		p.fx+=0.001*(mass*(cm_x-p.x)/(distance2*distance));
		p.fy+=0.001*(mass*(cm_y-p.y)/(distance2*distance));
	}else if (!isLeafNode()){
		for (unsigned int i=0;i<4;i++)
			children[i]->gatherForces(p);
	}else{
		iterate=data.getBegin();
		for(;iterate!=data.getEnd();iterate++){
			distance2=(iterate->x -p.x)*(iterate->x -p.x) + (iterate->y -p.y)*(iterate->y-p.y);
			distance=sqrt(distance2);
			if (distance>0.001){
				p.fx+=0.001*(mass*(cm_x-p.x)/(distance2*distance));
				p.fy+=0.001*(mass*(cm_y-p.y)/(distance2*distance));
			}
		}
	}
}

void QuadTree::assignForces(QuadTree* root){
	if (isLeafNode()){
		list<Point>::iterator iterate=data.getBegin();
		for(;iterate!=data.getEnd();iterate++){
			root->gatherForces(*iterate);
		}
	}else{
		for (unsigned int i=0;i<4;i++)
			children[i]->assignForces(root);
	}
}

void QuadTree::moveObjects(){
	if (isLeafNode()){
		list<Point>::iterator iterate=data.getBegin();
		for(;iterate!=data.getEnd();iterate++){
			iterate->vx+=iterate->fx;
			iterate->vy+=iterate->fy;

			iterate->x+=iterate->vx;
			iterate->y+=iterate->vy;
		}
	}else{
		for (unsigned int i=0;i<4;i++)
			children[i]->moveObjects();
	}
}

void QuadTree::reorder(QuadTree* root){
	static list<Point>::iterator iterate;
	static list<Point>::iterator delIterate;

	if (isLeafNode()){
		iterate=data.getBegin();
		for(;iterate!=data.getEnd();){
			if (!boundsTest(*iterate)){
				delIterate=iterate;
				iterate++;
				root->addPoint(*delIterate);
				data.removePoint(delIterate);
			}else{
				iterate++;
			}
		}
	}else{
		for (unsigned int i=0;i<children.size();i++)
			children[i]->reorder(root);
	}
}

void QuadTree::colasce(){//combine children
	if (isLeafNode()) return;

	if (mass < maxPoints){
		collectPoints(data);
		for (int i=0;i<4;i++){
			children[i]->deleteSelf(true);
		}
		children.clear();
	}else{
		for (int i=0;i<4;i++){
			children[i]->colasce();
		}
	}
}




//these 4 denote the quadrant masks.
unsigned char QuadTree::Q_ONE=1;
unsigned char QuadTree::Q_TWO=2;
unsigned char QuadTree::Q_THREE=4;
unsigned char QuadTree::Q_FOUR=8;

unsigned int QuadTree::maxDepth=64;
unsigned int QuadTree::maxPoints=4;
